/***********************************************************************/
/* File : Graph Output */

#include <stdio.h>

#include "__slam_level_type.h"
#include "__slam_level.h"
#include "__slam_tag.h"
#include "__slam_list.h"

/***********************************************************************/
/* External Function */

extern void __slam_graph_dump(
    void)
{
    FILE *graph_file;
    __slam_boolean_t to_next;

    graph_file = fopen("digraph.dot", "w");

    fprintf(graph_file, "digraph lattice {\n");
    LOOP_ALL_LIST_NODE(__slam_level_t, level, __slam_level_manager) {
	if (level == __slam_level_top ||
	    level == __slam_level_bottom) {
	    continue;
	}

	LOOP_ALL_LIST_NODE(__slam_level_t, target_level,
			   __slam_level_manager) {
	    if (target_level == __slam_level_top ||
		target_level == __slam_level_bottom) {
		continue;
	    }
	    if (__slam_level_eq(target_level, level)) {
		continue;
	    }
	    if (!__slam_level_leq(target_level, level)) {
		continue;
	    }

	    to_next = FALSE;
	    LOOP_ALL_LIST_NODE(__slam_level_t, check_level,
			       __slam_level_manager) {
		if (check_level == __slam_level_top ||
		    check_level == __slam_level_bottom) {
		    continue;
		}
		if (__slam_level_eq(check_level, level)) {
		    continue;
		}
		if (__slam_level_eq(target_level, check_level)) {
		    continue;
		}
		if (__slam_level_leq(target_level, check_level) &&
		    __slam_level_leq(check_level, level)) {

		    to_next = TRUE;
		    break;
		}
	    } LOOP_END;

	    if (!to_next) {

		fprintf(graph_file, "\"%s:", __slam_string_of_level(level));
		LOOP_ALL_LIST_NODE(__slam_tag_t, tag, level->lattice) {
		    __slam_tag_name_dump(graph_file, tag);
		} LOOP_END;

		fprintf(graph_file, "\" -> \"");

		fprintf(graph_file, "%s:", __slam_string_of_level(target_level));
		LOOP_ALL_LIST_NODE(__slam_tag_t, tag,
				   target_level->lattice) {
		    __slam_tag_name_dump(graph_file, tag);
		} LOOP_END;
		fprintf(graph_file, "\";\n");
	    }
	} LOOP_END;
    } LOOP_END;
    fprintf(graph_file, "}\n");

    fclose(graph_file);
}
