/***********************************************************************/
/* File : level */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>

#include "__slam_level.h"
#include "__slam_tag.h"
#include "__slam_list.h"

/***********************************************************************/
/* Define Macro. */

#define HASH_TABLE_SIZE	127

/***********************************************************************/
/* Data struct */

//! HashѤΥơ֥
typedef struct level_hash_data {
    int				 key;
    __slam_level_t		 level;
    struct level_hash_data	*next;
} level_hash_data_t;

//! Hashơ֥η
typedef level_hash_data_t	*level_hash_table_t;

/***********************************************************************/
/* Static Variable */

//! level ID ѤΥ󥿡
static unsigned int	 level_counter;

//! level ѥϥåơ֥
static level_hash_table_t	*level_hash_table;

/***********************************************************************/
/* Global Variable */

//! Ƥlevellist
__slam_list_t		*__slam_level_manager;

//! level  top
__slam_level_t		__slam_level_top;

//! level  bottom
__slam_level_t		__slam_level_bottom;

#ifndef NDEBUG
//! ǥХåѤΥե
FILE    *debug_file;
#endif

/***********************************************************************/
/* Static Function */

static __slam_level_t level_new(
    void);
static void level_destroy(
    __slam_level_t	level);

static void simply_lattice(
    __slam_list_t	*list);

static void level_lattice_list_to_array(
    __slam_level_t	level);

#ifndef NOT_USE_CACHE
static __slam_cache_table_t *cache_create(
    void);
static void cache_destroy(
    __slam_cache_table_t *cache_table);
static void cache_enter(
    __slam_cache_table_t	*cache_table,
    __slam_level_t	level,
    __slam_boolean_t	result);
static __slam_boolean_t cache_find(
    __slam_cache_table_t	*cache_table,
    __slam_level_t	level);
#endif

static level_hash_table_t *level_hash_create(
    void);
static void level_hash_flush(
    level_hash_table_t *hash_table);
static void level_hash_destroy(
    level_hash_table_t *hash_table);
static void level_hash_enter(
    level_hash_table_t	*hash_table,
    __slam_level_t	 level);
static __slam_level_t level_hash_find(
    level_hash_table_t	*hash_table,
    __slam_level_t	level);

/***********************************************************************/
/* External Function */

//!  
extern void __slam_level_init(
    void)
{
    level_counter = 0;
    __slam_level_manager = __slam_list_new();
    level_hash_table = level_hash_create();

    __slam_level_top = level_new();
    __slam_level_bottom = level_new();
}

//! 
extern void __slam_level_term(
    void)
{
#ifndef NDEBUG
    __slam_level_all_dump(debug_file);
#endif
    LOOP_ALL_LIST_NODE_MOD(__slam_level_t, level, __slam_level_manager) {
	level_destroy(level);
    } LOOP_END;
    level_hash_destroy(level_hash_table);
    __slam_list_delete(__slam_level_manager);
}

//! tag  level ֤
extern __slam_level_t __slam_level_of_tag(
    __slam_tag_t	tag)
{
    __slam_level_t	level;
    
/* ƱΤäƤ⿷˺ */
    level = level_new();
    __slam_list_insert(level->lattice, tag);

    level->num_of_tag = 1;
#ifdef TO_ARRAY
    level->lattice_array = calloc(1, sizeof(__slam_tag_t));
    level->lattice_array[0] = tag;
#endif
    level_hash_enter(level_hash_table, level);

    return level;
}

/*!
  ʣ tag  level֤(tag1ĤǤɤ)

  κǸˤϽüΰȤNULLĤ뤳ȡ
  NULLϡNULLʹߤ tag ̵뤵졢
  NULLޤǤ tag  level 롣
*/
extern __slam_level_t __slam_level_of_tags(
    __slam_tag_t	tag1, ...)
{
    __slam_tag_t	tag;
    __slam_level_t	level;
    va_list		ap;

    // level κ
    level = level_new();

    // tag(List)κ
    __slam_list_insert(level->lattice, tag1);

    va_start(ap, tag1);
    while((tag = va_arg(ap, __slam_tag_t)) != NULL) {
	__slam_list_insert(level->lattice, tag);
    }
    va_end(ap);

    level_lattice_list_to_array(level);

    level_hash_enter(level_hash_table, level);

    return level;
}

//! tag  level ֤
/*!
  Ʊ level ¸ߤС֤
*/
extern __slam_level_t __slam_level_of_tag_hash(
    __slam_tag_t	tag)
{
    __slam_level_t	level;
    __slam_level_t	ret_level;
    
    level = __slam_level_of_tag(tag);

    /* ƱΤϤ֤ */
    ret_level = level_hash_find(level_hash_table, level);
    if (ret_level) {
	level_destroy(level);
    }
    else {
	level_hash_enter(level_hash_table, level);
	ret_level = level;
    }

    return ret_level;
}

//! ʣ tag  level ֤(tag1ĤǤɤ)

/*!
  κǸˤϽüΰȤNULLĤ뤳ȡ
  NULLϡNULLʹߤ tag ̵뤵졢
  NULLޤǤ tag  level 롣

  Ʊ level ¸ߤС֤
*/
extern __slam_level_t __slam_level_of_tags_hash(
    __slam_tag_t	tag1, ...)
{
    __slam_tag_t	tag;
    __slam_level_t	level;
    __slam_level_t	ret_level;
    va_list		ap;

    // level κ
    level = level_new();

    // tag(List)κ
    __slam_list_insert(level->lattice, tag1);

    va_start(ap, tag1);
    while((tag = va_arg(ap, __slam_tag_t)) != NULL) {
	__slam_list_insert(level->lattice, tag);
    }
    va_end(ap);

    level_lattice_list_to_array(level);

    ret_level = level_hash_find(level_hash_table, level);
    if (ret_level) {
	level_destroy(level);
    }
    else {
	level_hash_enter(level_hash_table, level);
	ret_level = level;
    }

    return ret_level;
}

//! tag ν level κ
/*!
  va_list Ǥ tags νü NULL ǤʤФʤʤ
  va_list NULLϡNULL  tag  level
  롣
*/
extern __slam_level_t __slam_level_of_tags_va(
    va_list	tags)
{
    __slam_level_t	level;
    __slam_tag_t	tag;

    // level κ
    level = level_new();

    while((tag = va_arg(tags, __slam_tag_t)) != NULL) {
	__slam_list_insert(level->lattice, tag);
    }

    level_lattice_list_to_array(level);

    return level;
}

//! tag ν level κ
/*!
  va_list Ǥ tags νü NULL ǤʤФʤʤ
  va_list NULLϡNULL  tag  level
  롣
  Ʊ level ¸ߤС֤
*/
extern __slam_level_t __slam_level_of_tags_va_hash(
    va_list	tags)
{
    __slam_level_t	level;
    __slam_level_t	ret_level;
    __slam_tag_t	tag;

    // level κ
    level = level_new();

    while((tag = va_arg(tags, __slam_tag_t)) != NULL) {
	__slam_list_insert(level->lattice, tag);
    }

    level_lattice_list_to_array(level);

    ret_level = level_hash_find(level_hash_table, level);
    if (ret_level) {
	level_destroy(level);
    }
    else {
	level_hash_enter(level_hash_table, level);
	ret_level = level;
    }

    return level;
}

//! level ֤½֤
extern __slam_level_t __slam_level_or(
    __slam_level_t	level1,
    __slam_level_t	level2)
{
    __slam_level_t	level;

    if (level1 == level2) {
	return level1;
    }
    else if (level1 == __slam_level_bottom ||
	     level2 == __slam_level_top) {
	return level2;
    }
    else if (level2 == __slam_level_bottom ||
	     level1 == __slam_level_top) {
	return level1;
    }
    else {
	level = level_new();
	free(level->lattice);
	level->lattice = __slam_list_union(level1->lattice,
					   level2->lattice);
	
	level_lattice_list_to_array(level);

	return level;
    }
}

//!  ٥֤å
/*!
  - level1 <= level2 : True
  - otherwise        : False
*/
extern __slam_boolean_t __slam_level_leq(
    __slam_level_t	level1,
    __slam_level_t	level2)
{
    __slam_boolean_t	ret_val;
    __slam_boolean_t	is_found;
#ifdef TO_ARRAY
    int			i,j;
#endif
    assert(level1);
    assert(level2);

    if ((level1 == level2) ||
	(level1 == __slam_level_bottom) ||
	(level2 == __slam_level_top)) {
	return TRUE;
    }
    else if ((level1 == __slam_level_top) ||
	     (level2 == __slam_level_bottom)) {
	return FALSE;
    }
#ifndef NOT_USE_CACHE
    ret_val = cache_find(level2->result_cache_leq, level1);
    if (ret_val != 0) {
	return ret_val;
    }
#endif
#ifdef TO_ARRAY
    for(i = 0; i < level1->num_of_tag; i += 1) {
	is_found = FALSE;
	for(j = 0; j < level2->num_of_tag; j += 1) {
	    if (level1->lattice_array[i] ==
		level2->lattice_array[j]) {
		is_found = TRUE;
		break;
	    }
	    else if (__slam_list_member(level2->lattice_array[j]->small_set,
					level1->lattice_array[i])) {
		is_found = TRUE;
		break;
	    }
	}
	if (!is_found) {
	    return FALSE;
	}
    }
#else
    LOOP_ALL_LIST_NODE(__slam_tag_t, tag1, level1->lattice) {
	is_found = FALSE;
	LOOP_ALL_LIST_NODE(__slam_tag_t, tag2, level2->lattice) {
	    if (tag1 == tag2) {
		is_found = TRUE;
		break;
	    }
	    else if (__slam_list_member(tag2->small_set, tag1)) {
		is_found = TRUE;
		break;
	    }
	} LOOP_END;
	if (!is_found) {
	    return FALSE;
	}
    } LOOP_END;
#endif
    ret_val = TRUE;
#ifndef NOT_USE_CACHE
    /* TRUE Τߥå夹 */
    cache_enter(level2->result_cache_leq, level1, ret_val);
#endif
    return ret_val;
}

//!  ٥֤Ʊå
/*!
  - level1 = level2 : True
  - otherwise       : False
*/
extern __slam_boolean_t __slam_level_eq(
    __slam_level_t	level1,
    __slam_level_t	level2)
{
#ifdef TO_ARRAY
    int			i;
#endif
    assert(level1);
    assert(level2);

    if (level1 == level2 || level1->lattice == level2->lattice) {
	return TRUE;
    }
    else if ((level1 == __slam_level_top) ||
	     (level2 == __slam_level_top) ||
	     (level1 == __slam_level_bottom) ||
	     (level2 == __slam_level_bottom)) {
	return FALSE;
    }

    if (level1->num_of_tag != level2->num_of_tag) {
	return FALSE;
    }

#ifdef TO_ARRAY
    for(i = 0; i < level1->num_of_tag; i += 1) {
	if (level1->lattice_array[i] !=
	    level2->lattice_array[i]) {
	    return FALSE;
	}
    }

    // level1  level2 ƱʤΤlevel2Ƥä
    free(level2->lattice_array);
    level2->lattice_array = level1->lattice_array;
#else
    if (__slam_list_include(level1->lattice, level2->lattice) != 0 ){
      fprintf(stderr, "X\n");
      return FALSE;
    }
#endif

#ifndef NOT_USE_CACHE
    cache_destroy(level2->result_cache_leq);
    level2->result_cache_leq = level1->result_cache_leq;
#endif
    __slam_list_delete(level2->lattice);
    level2->lattice = level1->lattice;
    level2->is_same = TRUE;

    return TRUE;
}

// ǥХå
//! level˴ޤޤTagν()֤
//! üNULL
extern __slam_tag_t *__slam_tags_of_level(
    __slam_level_t	level)
{
    __slam_tag_t *ret_tags;
    int i;

    ret_tags = (__slam_tag_t *)calloc(level->num_of_tag + 1,
				      sizeof(__slam_tag_t));
    i = 0;
    LOOP_ALL_LIST_NODE(__slam_tag_t, tag, level->lattice) {
	ret_tags[i] = tag;
	i++;
    } LOOP_END;
    ret_tags[i] = NULL;

    return ret_tags;
}

//! levelμ֤̾
extern char *__slam_string_of_level(
    __slam_level_t	level)
{
    char	*string;

    string = (char *)malloc(sizeof("level000"));
    sprintf(string, "level%03d", level->id);

    return string;
}

//! levelΤƤtag֤̾
extern char *__slam_string_of_level_tags(
    __slam_level_t	level)
{
    char	*string;
    int		counter;
    
    // ɬפΰ
    counter = 0;
    LOOP_ALL_LIST_NODE(__slam_tag_t, tag, level->lattice) {
	counter += strlen(tag->name) + 1;
    } LOOP_END;

    // ɬפΰγ
    string = (char *)malloc(counter);

    // ʸν
    counter = 0;
    LOOP_ALL_LIST_NODE(__slam_tag_t, tag, level->lattice) {
	sprintf(&string[counter], "%s ", tag->name);
	counter += strlen(tag->name) + 1;
    } LOOP_END;

    return string;
}

//! level˴ޤޤ tag Ϥ롣
extern void __slam_level_dump(
    FILE		*file,
    __slam_level_t	level)
{
    fprintf(file, "level %3d: ", level->id);

    fprintf(file, "(");
    LOOP_ALL_LIST_NODE(__slam_tag_t, tag, level->lattice) {
	__slam_tag_name_dump(file, tag);
    } LOOP_END;

    fprintf(file, ")\n");
}

//! ƤlevelϤ롣
extern void __slam_level_all_dump(
    FILE	*file)
{
    fprintf(file, "------ All level Dump ----\n");
    LOOP_ALL_LIST_NODE(__slam_level_t, level, __slam_level_manager) {
	__slam_level_dump(file, level);
    } LOOP_END;
    fprintf(file, "--------------------------\n");
}

//!  level κ
static __slam_level_t level_new(
    void)
{
    __slam_level_t	level;

    level = (__slam_level_t)malloc(sizeof(__slam_level_type));
    level->id          = level_counter;
    level->lattice = __slam_list_new();
#ifdef TO_ARRAY
    level->lattice_array = NULL;
#endif
#ifndef NOT_USE_CACHE
    level->result_cache_leq = cache_create();
    __slam_list_insert(__slam_level_manager, level);
#endif
    level->is_same = FALSE;

    level_counter += 1;

    return level;
}

//! level ξõ
static void level_destroy(
    __slam_level_t	level)
{
    if (!level->is_same) {
#ifndef NOT_USE_CACHE
	cache_destroy(level->result_cache_leq);
#endif
#ifdef TO_ARRAY
	free(level->lattice_array);
#endif
	__slam_list_delete(level->lattice);
    }
    __slam_list_extract(__slam_level_manager, level);
    free(level);

    return;
}

//! level  tag δά
static void simply_lattice(
    __slam_list_t	*list)
{
    LOOP_ALL_LIST_NODE(__slam_tag_t, tag, list) {
	__slam_list_sub(list, tag->small_set);
	__slam_list_insert(list, tag);
    } LOOP_END;
}

#ifndef NOT_USE_CACHE
//!  åΰݤ롣
/*!
 level ֤ӷ̤򥭥å뤿
 åΰݤ롣
 ӷ̤ leq  TRUE Ǥ롣
 FALSEξ VITCǤ Type Error ʤΤϿƤ
 ̵̤Ǥꡢ
 ޤFALSE(0)֤ȡå̤Ͽ(0)
 ̤ǤʤΤFALSEϿʤ
 ʳϿʤ
*/
static __slam_cache_table_t *cache_create(
    void)
{
    __slam_cache_table_t *cache_table;

    cache_table = (__slam_cache_table_t *)calloc(CACHE_TABLE_SIZE,
						 sizeof(__slam_cache_data_t *));
    if (cache_table == NULL) {
	printf("calloc error\n");
    }
    return cache_table;
}

//! åեå夹롣
/*!
  ˲Ū rule ɲûѤ롣
*/
extern void __slam_cache_flush(
    __slam_cache_table_t *cache_table)
{
    int i;
    __slam_cache_data_t	*cur_hash_data;
    __slam_cache_data_t	*next_hash_data;

    for(i = 0; i < CACHE_TABLE_SIZE; i++) {
	cur_hash_data = cache_table[i];
	while (cur_hash_data) {
	    next_hash_data = cur_hash_data->next;
	    free(cur_hash_data);
	    cur_hash_data = next_hash_data;
	}
    }
    return;
}

//! åξõ
static void cache_destroy(
    __slam_cache_table_t *cache_table)
{
    __slam_cache_flush(cache_table);
    free(cache_table);
    return;
}

//! åӷ̤Ͽ
/*!
  leq  ̤ TRUE ξΤϿ롣
*/
static void cache_enter(
    __slam_cache_table_t	*cache_table,
    __slam_level_t	level,
    __slam_boolean_t	result)
{
    __slam_cache_data_t	*new_hash_data;
    __slam_cache_data_t	*cur_hash_data;
    __slam_cache_data_t	*prev_hash_data;
    int			key;

    key = (int)level;

    cur_hash_data = cache_table[key % CACHE_TABLE_SIZE];
    prev_hash_data = NULL;
    while (cur_hash_data) {
	if (cur_hash_data->key < key) {
	    break;
	}
	else if (cur_hash_data->key == key) {
	    /* ϿѤ */
	    return;
	}
	prev_hash_data = cur_hash_data;
	cur_hash_data = cur_hash_data->next;
    }

    /* ơ֥ */
    new_hash_data = (__slam_cache_data_t *)malloc(sizeof(__slam_cache_data_t));
    new_hash_data->key = key;
    new_hash_data->result = result;
    new_hash_data->next = cur_hash_data;

    if (prev_hash_data == NULL) {
	cache_table[key % CACHE_TABLE_SIZE] = new_hash_data;
    }
    else {
	prev_hash_data->next = new_hash_data;
    }
    return;
}

//! åӷ̤򸡺
/*!
  åϿƤTRUE֤
  åˤʤϡFALSE֤
  åϿƤͤTRUEΤߤǡ
  FALSE̤Ͽɽ
*/
static __slam_boolean_t cache_find(
    __slam_cache_table_t	*cache_table,
    __slam_level_t		level)
{
    __slam_cache_data_t	*hash_data;
    int			 key;

    key = (int)level;

    hash_data = cache_table[key % CACHE_TABLE_SIZE];

    while (hash_data) {
	if (hash_data->key == key) {
	    /* ȯ */
	    return hash_data->result;
	}
	else if (hash_data->key < key) {
	    /* ϥåˤʤ */
	    break;
	}
	else {
	    /* õ */
	    hash_data = hash_data->next;
	}
    }
    return FALSE;
}
#endif

//! ϥåơ֥κ
/*!
  level ֤ƱΤϡƱ level֤褦
  ϥåϿ롣
*/
static level_hash_table_t *level_hash_create(
    void)
{
    level_hash_table_t *hash_table;

    hash_table = (level_hash_table_t *)calloc(HASH_TABLE_SIZE,
					      sizeof(level_hash_data_t *));
    if (hash_table == NULL) {
	printf("calloc error\n");
    }
    return hash_table;
}

//! levelΥϥåեå夹
static void level_hash_flush(
    level_hash_table_t *hash_table)
{
    int i;
    level_hash_data_t	*cur_hash_data;
    level_hash_data_t	*next_hash_data;

    for(i = 0; i < HASH_TABLE_SIZE; i++) {
	cur_hash_data = hash_table[i];
	while (cur_hash_data) {
	    next_hash_data = cur_hash_data->next;
	    free(cur_hash_data);
	    cur_hash_data = next_hash_data;
	}
    }
    return;
}

//! level Υϥåõ롣
static void level_hash_destroy(
    level_hash_table_t *hash_table)
{
    level_hash_flush(hash_table);
    free(hash_table);
    return;
}

//! level ϥåϿ롣
static void level_hash_enter(
    level_hash_table_t	*hash_table,
    __slam_level_t	 level)
{
    level_hash_data_t	*new_hash_data;
    level_hash_data_t	*cur_hash_data;
    level_hash_data_t	*prev_hash_data;
    int			 key;

    key = level->num_of_tag;

    cur_hash_data = hash_table[key % HASH_TABLE_SIZE];
    prev_hash_data = NULL;
    while (cur_hash_data) {
	if (cur_hash_data->key < key) {
	    break;
	}
	else if (cur_hash_data->key == key &&
		 __slam_level_eq(cur_hash_data->level, level)) {
	    /* ϿѤ */
	    return;
	}
	prev_hash_data = cur_hash_data;
	cur_hash_data = cur_hash_data->next;
    }

    /* ơ֥ */
    new_hash_data = (level_hash_data_t *)malloc(sizeof(level_hash_data_t));
    new_hash_data->key = key;
    new_hash_data->level = level;
    new_hash_data->next = cur_hash_data;

    if (prev_hash_data == NULL) {
	hash_table[key % HASH_TABLE_SIZE] = new_hash_data;
    }
    else {
	prev_hash_data->next = new_hash_data;
    }
    return;
}

//! ϥå夫level 򸫤Ĥ롣
/*!
  ˡϡޤleveltagοǿʬ
  __slam_level_eq()Ĵ٤롣
*/
static __slam_level_t level_hash_find(
    level_hash_table_t	*hash_table,
    __slam_level_t	 level)
{
    level_hash_data_t	*hash_data;
    int			 key;

    key = level->num_of_tag;

    hash_data = hash_table[key % HASH_TABLE_SIZE];

    while (hash_data) {
	if (hash_data->key == key &&
	    __slam_level_eq(hash_data->level, level)) {
	    /* ȯ */
	    return hash_data->level;
	}
	else if (hash_data->key < key) {
	    /* ϥåˤʤ */
	    break;
	}
	else {
	    /* õ */
	    hash_data = hash_data->next;
	}
    }
    return NULL;
}

//! level  tag lattice  list¤  Ѵ
static void level_lattice_list_to_array(
    __slam_level_t	level)
{
#ifdef TO_ARRAY
    int i;
#endif

    // tagδά
    simply_lattice(level->lattice);

    // ο
    level->num_of_tag = level->lattice->length;
#ifdef TO_ARRAY
    // Ǽΰ
    level->lattice_array = calloc(level->num_of_tag,
				  sizeof(__slam_tag_t));

    // Ǽ
    i = 0;
    LOOP_ALL_LIST_NODE(__slam_tag_t, tag, level->lattice) {
	level->lattice_array[i] = tag;
	i += 1;
    } LOOP_END;

    assert(i == level->num_of_tag);
#endif
    return;
}
