/***********************************************************************/
/* File : ѥꥹȽ */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "__slam_list.h"
#include "__slam_boolean.h"

#define CHECKPOINT printf("line : %04d in %s\n", __LINE__, __FILE__);

/***********************************************************************/
/* Static Variable */

#ifdef USE_FREE_LIST_NODE
static __slam_list_t *list_free=NULL;
#endif

/***********************************************************************/
/* Static Function decl. */
static __slam_list_node_t *list_node_new(
    void);
static void list_node_destroy(
    __slam_list_node_t	*list_node);
static void list_node_delete(
    __slam_list_node_t	*list);

/***********************************************************************/
/* Function */
extern void __slam_list_init(
    void)
{
#ifdef USE_FREE_LIST_NODE
    list_free = __slam_list_new();
#endif
}

extern void __slam_list_term(
    void )
{
#ifdef USE_FREE_LIST_NODE
    __slam_list_destroy(list_free);
#endif
}

extern __slam_list_t *__slam_list_new(
    void)
{
    __slam_list_t	*list;

    list = (__slam_list_t *)malloc(sizeof(__slam_list_t));
    list->head = NULL;
    list->length = 0;

    return list;
}

extern void __slam_list_delete(
    __slam_list_t	*list)
{
    __slam_list_node_t	*cur_node;
    __slam_list_node_t	*next_node;

    assert(list);

    cur_node = list->head;
    while (cur_node) {
	next_node = (cur_node != NULL) ? cur_node->next : NULL;
	list_node_delete(cur_node);
	cur_node = next_node;
    }

    free(list);

    return;
}

extern void __slam_list_destroy(
    __slam_list_t	*list)
{
    __slam_list_node_t	*cur_node;
    __slam_list_node_t	*next_node;

    assert(list);

    cur_node = list->head;
    while (cur_node) {
	next_node = (cur_node != NULL) ? cur_node->next : NULL;
	list_node_destroy(cur_node);
	cur_node = next_node;
    }

    free(list);

    return;
}

static __slam_list_node_t *list_node_new(
    void)
{
    __slam_list_node_t *list_node;

#ifdef USE_FREE_LIST_NODE
    if (list_free->head != NULL) {
	list_node = list_free->head;
	list_free->head = list_node->next;
	list_free->length = list_free->length - 1;

	assert(list_free->length >= 0);

	list_node->next = NULL;
	list_node->data = NULL;

	return list_node;
    }
#endif
    list_node = (__slam_list_node_t *)malloc(sizeof(__slam_list_node_t));
    list_node->next = NULL;
    list_node->data = NULL;

    return list_node;
}

static void list_node_delete(
    __slam_list_node_t	*list_node)
{
    assert(list_node);
#ifdef USE_FREE_LIST_NODE
    list_node->next = list_free->head ;
    list_free->head = list_node;
    list_node->data = NULL;
    list_free->length = list_free->length + 1;
#else
    free(list_node);
#endif
    return;
}

static void list_node_destroy(
    __slam_list_node_t	*list_node)
{
    assert(list_node);
    free(list_node);
    return;
}

extern void __slam_list_insert(
    __slam_list_t	*list,
    void	*data)
{
    __slam_list_node_t	*node;
    __slam_list_node_t	*prev_node;
    __slam_list_node_t	*cur_node;

    assert(list);
    assert(data);

    prev_node = NULL;
    cur_node = list->head;
    while (cur_node) {
	if (cur_node->data < data) {
	    break;
	}
	else if (cur_node->data == data) {
	    /* ϿѤ */
	    return;
	}
	prev_node = cur_node;
	cur_node = cur_node->next;
    }

    node = list_node_new();
    node->next = cur_node;
    node->data = data;

    if (prev_node == NULL) {
	list->head = node;
    }
    else {
	prev_node->next = node;
    }

    list->length += 1;
}

extern void __slam_list_extract(
    __slam_list_t	*list,
    void	*data)
{
    __slam_list_node_t	*prev_node;
    __slam_list_node_t	*cur_node;

    assert(list);
    assert(data);

    prev_node = NULL;
    cur_node = list->head;
    while (cur_node) {
	if (cur_node->data < data) {
	    break;
	}
	else if (cur_node->data == data) {
	    /* оݤȯ */
	    if (prev_node == NULL) {
		list->head = cur_node->next;
	    }
	    else {
		prev_node->next = cur_node->next;
	    }
	    list->length -= 1;
	    return;
	}
	prev_node = cur_node;
	cur_node = cur_node->next;
    }
}

extern __slam_boolean_t __slam_list_member(
    __slam_list_t	*list,
    void		*data)
{
    __slam_list_node_t	*cur_node;

    assert(list);
    assert(data);

    cur_node = list->head;
    while (cur_node) {
	if (cur_node->data < data) {
	    return FALSE;
	}
	else if (cur_node->data == data) {
	    return TRUE;
	}
	cur_node = cur_node->next;
    }
    return FALSE;
}

/* list1  list2 ޤफ */
/*
  list1  list2 : > 0
  list1  list2 : = 0
  otherwise      : < 0
*/
extern int __slam_list_include(
    __slam_list_t	*list1,
    __slam_list_t	*list2)
{
    __slam_list_node_t	*node1;
    __slam_list_node_t	*node2;
    __slam_boolean_t	 list2_is_long;

    assert(list1);
    assert(list2);

    node1 = list1->head;
    node2 = list2->head;

    list2_is_long = FALSE;
    while (node1 && node2) {
	if (node2->data > node1->data) {
	    /* list2ˤälist1ˤϤʤ */
	    node2 = node2->next;
	    list2_is_long = TRUE;
	}
	else if (node2->data < node1->data) {
	    /* list2ʤlist1ˤ */
	    return -1;
	}
	else {
	    /* node1node2 Ʊǡ */
	    node1 = node1->next;
	    node2 = node2->next;
	}
    }

    if (node2) {
	/* list2˻Ĥ꤬ */
	return 1;
    }
    else if (node1) {
	/* list1˻Ĥ꤬ */
	return -1;
    }
    else if (list2_is_long) {
	/* list1˻Ĥ꤬ʤlist2ˤĤ꤬ʤ */
	/* list2 Ĺ TRUE */
	return 1;
    }
    else {
	/* list1˻Ĥ꤬ʤlist2ˤĤ꤬ʤ */
	/* list1ĹʤĤޤ list1 = list2 */
	return 0;
    }
}

/* list1  list2  Union ä֤ */
extern __slam_list_t *__slam_list_union(
    __slam_list_t	*list1,
    __slam_list_t	*list2)
{
    __slam_list_node_t	*node1;
    __slam_list_node_t	*node2;
    __slam_list_node_t	*prev_node;
    __slam_list_node_t	*new_node;
    __slam_list_t	*unioned_list;
 
    assert(list1);
    assert(list2);

    node1 = list1->head;
    node2 = list2->head;

    unioned_list = __slam_list_new();

    prev_node = NULL;
    while (node1 && node2) {
	/* Ρɤ */
	new_node = list_node_new();

	/* Ρɤͤ */	
	if (node1->data > node2->data) {
	    /* node1ɲ */
	    new_node->data = node1->data;
	    node1 = node1->next;
	}
	else if (node1->data < node2->data) {
	    /* node2ɲ */
	    new_node->data = node2->data;
	    node2 = node2->next;
	}
	else {
	    /* node1node2 Ʊǡ */
	    new_node->data = node1->data;
	    node1 = node1->next;
	    node2 = node2->next;
	}

	/* ΡɤꥹȤɲ */
	if (prev_node == NULL) {
	    unioned_list->head = new_node;
	}
	else {
	    prev_node->next = new_node;
	}
	prev_node = new_node;
	unioned_list->length += 1;
    }

    /* node1ʲɲ */
    while (node1) {
	new_node = list_node_new();
	new_node->data = node1->data;

	if (prev_node == NULL) {
	    unioned_list->head = new_node;
	}
	else {
	    prev_node->next = new_node;
	}
	prev_node = new_node;
	unioned_list->length += 1;

	node1 = node1->next;
    }

    /* node2ʲɲ */
    while (node2) {
	new_node = list_node_new();
	new_node->data = node2->data;
	
	if (prev_node == NULL) {
	    unioned_list->head = new_node;
	}
	else {
	    prev_node->next = new_node;
	}
	prev_node = new_node;
	unioned_list->length += 1;

	node2 = node2->next;
    }

    return unioned_list;
}

/* list1  list2 ɲ  */
extern void __slam_list_add(
    __slam_list_t	*list1,
    __slam_list_t	*list2)
{
    __slam_list_node_t	*node1;
    __slam_list_node_t	*node2;
    __slam_list_node_t	*prev_node;
    __slam_list_node_t	*new_node;

    assert(list1);
    assert(list2);

    node1 = list1->head;
    node2 = list2->head;

    prev_node = NULL;
    while (node1 && node2) {
	if (node1->data > node2->data) {
	    prev_node = node1;
	    node1 = node1->next;
	}
	else if (node1->data < node2->data) {
	    /* Ρɤ */
	    new_node = list_node_new();

	    /* node2ɲ */
	    new_node->data = node2->data;

	    if (prev_node == NULL) {
		list1->head = new_node;
	    }
	    else {
		prev_node->next = new_node;
	    }
	    new_node->next = node1;
	    prev_node = new_node;

	    node2 = node2->next;
	    list1->length += 1;
	}
	else {
	    prev_node = node1;
	    node1 = node1->next;
	    node2 = node2->next;
	}
    }

    /* node2ʲɲ */
    while (node2) {
	new_node = list_node_new();
	new_node->data = node2->data;
	
	if (prev_node == NULL) {
	    list1->head = new_node;
	}
	else {
	    prev_node->next = new_node;
	}
	prev_node = new_node;
	list1->length += 1;

	node2 = node2->next;
    }
}

/* list1  list2  */
extern void __slam_list_sub(
    __slam_list_t	*list1,
    __slam_list_t	*list2)
{
    __slam_list_node_t	*node1;
    __slam_list_node_t	*node2;
    __slam_list_node_t	*prev_node;

    assert(list1);
    assert(list2);

    node1 = list1->head;
    node2 = list2->head;

    prev_node = NULL;
    while (node1 && node2) {
	if (node1->data > node2->data) {
	    prev_node = node1;
	    node1 = node1->next;
	}
	else if (node1->data < node2->data) {
	    node2 = node2->next;
	}
	else {
	    if (prev_node == NULL) {
		list1->head = node1->next;
		list_node_delete(node1);
		node1 = list1->head;
	    }
	    else {
		prev_node->next = node1->next;
		list_node_delete(node1);
		node1= prev_node->next;
	    }
	    node2 = node2->next;
	    list1->length -= 1;
	}
    }

    return;
}

/***********************************************************************/
/* For Debug */
extern void __slam_list_dump(
    FILE		*file,
    __slam_list_t	*list)
{
    __slam_list_node_t	*node;
    int counter;

    fprintf(file, "List:");

    counter = 1;
    node = list->head;
    while (node) {
	fprintf(file, "\t0x%x", (unsigned int)node->data);
	node = node->next;
	if (counter > 5) {
	    fprintf(file, "\n");
	    counter++;
	}
	else {
	    counter = 1;
	}
    }
    fprintf(file, "\n");
}
