/***********************************************************************/
/* File : Rule */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "__slam_list.h"
#include "__slam_tag.h"
#include "__slam_rule.h"
#include "__slam_level.h"

/***********************************************************************/
/* Static Variable */

#ifndef NDEBUG
static unsigned int	 rule_counter;
static __slam_list_t	*rule_manager;
#endif

/***********************************************************************/
/* Global Variable */

/***********************************************************************/
/* Static Function */
#ifndef NDEBUG
static void rule_destroy(
    __slam_rule_t	rule);
#endif

/***********************************************************************/
/* External Function */

/*  */
extern void __slam_rule_init(
    void)
{
#ifndef NDEBUG
    rule_counter = 0;
    rule_manager  = __slam_list_new();
#endif
}

/*  */
extern void __slam_rule_term(
    void)
{
#ifndef NDEBUG
    __slam_rule_all_dump(debug_file);
    LOOP_ALL_LIST_NODE(__slam_rule_t, rule, rule_manager) {
	rule_destroy(rule);
    } LOOP_END;
    __slam_list_delete(rule_manager);
#endif
}

/*!
  Ruleɲ

  small_tag =< big_tag  rule 롣
  small_tag  big_tagƱtagξϤʤˤ⤷ʤ
*/
extern void __slam_add_rule(
    __slam_tag_t	small_tag,
    __slam_tag_t	big_tag)
{
#ifndef NDEBUG
    __slam_rule_t	rule;
#endif
    __slam_list_t	*small_tag_set;
    __slam_list_t	*big_tag_set;

    assert(big_tag);
    assert(small_tag);

    if (__slam_tag_eq(small_tag, big_tag)) {
	return;
    }

#ifndef NDEBUG
    rule = (__slam_rule_t)malloc(sizeof(__slam_rule_type));
    rule->id = rule_counter;
    rule->big = big_tag;
    rule->small = small_tag;

    __slam_list_insert(rule_manager, rule);

    rule_counter += 1;
#endif

    small_tag_set = __slam_list_new();
//    __slam_list_insert(small_tag_set, small_tag);
    __slam_list_add(small_tag_set, small_tag->small_set);

    __slam_list_add(big_tag->small_set, small_tag_set);
    LOOP_ALL_LIST_NODE(__slam_tag_t, bigger_tag, big_tag->big_set) {
	__slam_list_add(bigger_tag->small_set, small_tag_set);
    } LOOP_END;

    __slam_list_delete(small_tag_set);

    big_tag_set = __slam_list_new();
    __slam_list_insert(big_tag_set, big_tag);
    __slam_list_add(big_tag_set, big_tag->big_set);

    __slam_list_add(small_tag->big_set, big_tag_set);
    LOOP_ALL_LIST_NODE(__slam_tag_t, smaller_tag, small_tag->small_set) {
	__slam_list_add(smaller_tag->big_set, big_tag_set);
    } LOOP_END;
    __slam_list_delete(big_tag_set);

#ifndef NDEBUG
    __slam_list_insert(big_tag->small_tags, small_tag);
#endif
    return;
}

/* ˲ŪRuleɲ */
extern void __slam_force_add_rule(
    __slam_tag_t	small_tag,
    __slam_tag_t	big_tag)
{
    __slam_add_rule(small_tag, big_tag);
#ifndef NOT_USE_CACHE
    LOOP_ALL_LIST_NODE(__slam_level_t, level, __slam_level_manager) {
	/* cache flush */
	__slam_cache_flush(level->result_cache_leq);
    } LOOP_END;
#endif
}

#ifndef NDEBUG
static void rule_destroy(
    __slam_rule_t	rule)
{
    free(rule);
}

/* ǥХå */
extern void __slam_rule_dump(
    FILE		*file,
    __slam_rule_t	rule)
{
#ifndef NDEBUG_USE_GRAPH
    FILE *graph_file;
#endif

    fprintf(file, "Rule  %3d: ", rule->id);
    __slam_tag_name_dump(file, rule->small);
    fprintf(file, "<= ");
    __slam_tag_name_dump(file, rule->big);
    fprintf(file, "\n");

#ifndef NDEBUG_USE_GRAPH
    graph_file = fopen("digraph_rule.dot", "w");
    fprintf(graph_file, "digraph tag_lattice {\n");
    LOOP_ALL_LIST_NODE(__slam_rule_t, rule, rule_manager) {
	fprintf(graph_file, "\"");
	__slam_tag_name_dump(graph_file, rule->big);
	fprintf(graph_file, "\" -> \"");
	__slam_tag_name_dump(graph_file, rule->small);
	fprintf(graph_file, "\";\n");
    } LOOP_END;
    fprintf(graph_file, "}\n");

    fclose(graph_file);
#endif
}

extern void __slam_rule_all_dump(
    FILE	*file)
{
    fprintf(file, "------ All Rule Dump -----\n");
    LOOP_ALL_LIST_NODE(__slam_rule_t, rule, rule_manager) {
	__slam_rule_dump(file, rule);
    } LOOP_END;
    fprintf(file, "--------------------------\n");
}
#endif
