/***********************************************************************/
/* File : Tag */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <string.h>
#include <assert.h>

#include "__slam_tag_type.h"
#include "__slam_tag.h"
#include "__slam_rule.h"
#include "__slam_level.h"
#include "__slam_list.h"

/***********************************************************************/
/* Define Macro. */

#define HASH_TABLE_SIZE		7

#ifndef NDEBUG
#define SLAM_PREFIX	""
#else
#define SLAM_PREFIX	"__slam_tag_"
#endif

/***********************************************************************/
/* Define type */
//! TagѤΥϥåơ֥
typedef struct tag_hash_data {
    char			*tag_name;
    __slam_tag_t		 tag;
    struct tag_hash_data	*next;
} tag_hash_data_t;

typedef tag_hash_data_t *tag_hash_table_t;

typedef struct tag_unix_hash_data {
    long long int		 uid_gid;
    __slam_boolean_t		 has_tags;
    struct tag_unix_hash_data	*next;
} tag_unix_hash_data_t;

typedef tag_unix_hash_data_t *tag_unix_hash_table_t;

/***********************************************************************/
/* Static Variable */

#ifndef NDEBUG
/* Tag ID(Tagο) */
static unsigned int		tag_counter;
#endif

/* Tag ѥꥹ */
static __slam_list_t		*tag_manager;

/* Tag ѥϥåơ֥ */
static tag_hash_table_t		*tag_hash_table;

/* Unix Tag ѥϥåơ֥ */
static tag_unix_hash_table_t	*tag_unix_hash_table;

/***********************************************************************/
/* Global Variable */

/* Unix Permmission Tag */
__slam_tag_t	__slam_tag_unix_root;
__slam_tag_t	__slam_tag_unix_all;

/***********************************************************************/
/* Static Function */
static __slam_tag_t tag_new(
    char	*tag_name);
static void tag_destroy(
    __slam_tag_t	tag);

static tag_hash_table_t *tag_hash_create(
    void);
static void tag_hash_flush(
    tag_hash_table_t *hash_table);
static void tag_hash_destroy(
    tag_hash_table_t *hash_table);
static void tag_hash_enter(
    tag_hash_table_t	*hash_table,
    char		*tag_name,
    __slam_tag_t	tag);
static  __slam_tag_t tag_hash_find(
    tag_hash_table_t	*hash_table,
    char		*name);

static tag_unix_hash_table_t *tag_unix_hash_create(
    void);
static void tag_unix_hash_flush(
    tag_unix_hash_table_t *hash_table);
static void tag_unix_hash_destroy(
    tag_unix_hash_table_t *hash_table);
static void tag_unix_hash_enter(
    tag_unix_hash_table_t	*hash_table,
    long long int		key,
    __slam_boolean_t		has_tag);
static  __slam_boolean_t tag_unix_hash_find(
    tag_unix_hash_table_t	*hash_table,
    long long int		key);
static __slam_boolean_t is_uid_member_of_gid(
    uid_t	uid,
    gid_t	gid);

static __slam_tag_t __slam_tag_unix_group_suser(
    uid_t	uid,
    gid_t	gid);
static __slam_tag_t __slam_tag_unix_group_auser(
    uid_t	uid,
    gid_t	gid);
static __slam_tag_t __slam_tag_unix_all_suser(
    uid_t	uid);
static __slam_tag_t __slam_tag_unix_all_sgroup(
    gid_t	gid);
static __slam_tag_t __slam_tag_unix_all_sgroup_suser(
    uid_t	uid,
    gid_t	gid);
static __slam_tag_t __slam_tag_unix_all_sgroup_auser(
    uid_t	uid,
    gid_t	gid);
static __slam_boolean_t tag_unix_add_to_hash(
    uid_t	uid,
    gid_t	gid);

/***********************************************************************/
/* External Function */

/*  */
extern void __slam_tag_init(
    void)
{
#ifndef NDEBUG
    tag_counter = 0;
#endif
    tag_manager = __slam_list_new();
    tag_hash_table = tag_hash_create();
    tag_unix_hash_table = tag_unix_hash_create();

    __slam_tag_unix_root = tag_new("__slam_tag_unix_root");
    __slam_list_insert(__slam_tag_unix_root->small_set,
		       __slam_tag_unix_root);
    __slam_tag_unix_root->is_unix = TRUE;
    __slam_tag_unix_root->kind = UNIX_ROOT;

    __slam_tag_unix_all = tag_new("__slam_tag_unix_all");
    __slam_list_insert(__slam_tag_unix_all->small_set,
		       __slam_tag_unix_all);
    __slam_tag_unix_all->is_unix = TRUE;
    __slam_tag_unix_root->kind = UNIX_ALL;
}

/*  */
extern void __slam_tag_term(
    void)
{
#ifndef NDEBUG
    __slam_tag_all_dump(debug_file);
#endif
    tag_unix_hash_destroy(tag_unix_hash_table);
    tag_hash_destroy(tag_hash_table);
    LOOP_ALL_LIST_NODE(__slam_tag_t, tag, tag_manager) {
	tag_destroy(tag);
    } LOOP_END;
    __slam_list_delete(tag_manager);
}

//! tagκ
/*!
  tag_nameǼ̤ˤФtag֤
  ʤп˺롣

  leqθѤ˼ʬȤsmall_set롣
*/
extern __slam_tag_t __slam_add_tag(
    char	*tag_name)
{
    __slam_tag_t	tag;

    tag = __slam_find_tag(tag_name);
    if (tag == NULL) {
	tag = tag_new(tag_name);
	__slam_list_insert(tag->small_set, tag);
    }

    return tag;
}

//! Tagκ ̾Υԡդ
/*!
  tag_nameǼ̤ˤФtag֤
  ʤп˺롣

  leqθѤ˼ʬȤsmall_set롣

  tag_name ƤӽФ free() ǽ
  ˻Ѥ롣
*/
extern __slam_tag_t __slam_add_tag_copy(
    char	*tag_name)
{
    __slam_tag_t	tag;
    int			tag_name_length;
    char		*tag_name_copy;

    tag = __slam_find_tag(tag_name);
    if (tag == NULL) {
	tag_name_length = strlen(tag_name);
	tag_name_copy = (char *)malloc(tag_name_length+1);
	strcpy(tag_name_copy, tag_name);
	tag = tag_new(tag_name_copy);
	__slam_list_insert(tag->small_set, tag);
    }

    return tag;
}

/* Tagθ */
extern __slam_tag_t __slam_find_tag(
    char	*tag_name)
{
    return tag_hash_find(tag_hash_table, tag_name);
}

/* RuleνطĴ */
/*
** tag1 <= tag2 : True
** tag1 >  tag2 : False
** طʤ : False
*/
extern __slam_boolean_t __slam_tag_leq(
    __slam_tag_t	tag1,
    __slam_tag_t	tag2)
{
    __slam_boolean_t	ret_val;

    if (tag1 == tag2) {
	ret_val = TRUE;
    }
    else {
	ret_val = __slam_list_member(tag2->small_set, tag1);
    }
    return ret_val;
}    

/* RuleνطĴ */
/*
** tag1 = tag2  : True
** tag1 <> tag2 : False
** طʤ : False
**
** ȡ˲Ūʥ롼ɲäˤꡢۤʤ륿Ǥ
** δطˤʤꤦ롣tag1 == tag2 ǤϤᡣ
** ޤ̾ۤʤΤ tag1 tag2Υݥ󥿤
** Ƥᡣ
*/
extern __slam_boolean_t __slam_tag_eq(
    __slam_tag_t	tag1,
    __slam_tag_t	tag2)
{
    __slam_boolean_t	ret_val;

    if (tag1 == tag2) {
	ret_val = TRUE;
    }
    else if (__slam_list_include(tag1->small_set, tag2->small_set) == 0) {
	ret_val = TRUE;
    }
    else {
	ret_val = FALSE;
    }
    return ret_val;
}    

//! Unix File System permissionˤtagκ
/*!
  - permission(mode)  user/groupδط tag
    -- --- : root
    -- r-- : user
    -- rr- : if (user  group) group
            else               user+group
    -- r-r : if (user  group) all-group+user
            else               all-group
    -- -r- : if (user  group) group-user
            else               group
    -- --r : if (user  group) all-group
            else               all-group-user
    -- -rr : all-user
    -- rrr : all

  - ʲϡuser  group ξ硢tag˼ưŪɲä롼
    -- all =< group
    -- all =< all-user
    -- all =< all-group+user
    -- group =< user
    -- group =< group-user
    -- all-group+user =< user
    -- all-group+user =< all-group
    -- all-user =< group-user
    -- all-user =< all-group
    -- user =< root
    -- group-user =< root
    -- all-group =< root

  - ʲϡuser ! group ξ硢tag˼ưŪɲä롼
    -- all =< group+user
    -- all =< all-user
    -- all =< all-group
    -- group+user =< user
    -- group+user =< group
    -- all-user =< group
    -- all-user =< all-group-user
    -- all-group =< user
    -- all-group =< all-group-user
    -- user =< root
    -- group =< root
    -- all-group-user =< root

  - group֤޴طĴ٤ʤ
    ĤޤꡢgroupۤʤоήȤƤϥ顼Ȥʤ
*/
extern __slam_tag_t __slam_tag_unix(
    uid_t	uid,
    gid_t	gid,
    mode_t	mode)
{
    __slam_tag_t	ret_tag;
    __slam_tag_t	tag_root;		/* root */
    __slam_tag_t	tag_user;		/* user */
    __slam_tag_t	tag_group;		/* group */
    __slam_tag_t	tag_all;		/* all */
    __slam_tag_t	tag_group_auser;	/* group + user */
    __slam_tag_t	tag_group_suser;	/* group - user */
    __slam_tag_t	tag_all_suser;		/* all - user */
    __slam_tag_t	tag_all_sgroup;		/* all - group */
    __slam_tag_t	tag_all_sgroup_suser;	/* all - group - user */
    __slam_tag_t	tag_all_sgroup_auser;	/* all - group + user */

    /* <uid, gid> ȤǤ tag 뤳ȤϿ */
    if (tag_unix_add_to_hash(uid, gid)) {
	/* UIDGID ˴ؤ tag ޤʤ硢tag  rule  */
	if (is_uid_member_of_gid(uid, gid)) {
	    /* UID  GID ˴ޤޤ */

	    /* tag  */
	    tag_root             = __slam_tag_unix_root;
	    tag_user             = __slam_tag_unix_user(uid);
	    tag_group            = __slam_tag_unix_group(gid);
	    tag_all              = __slam_tag_unix_all;
	    tag_group_suser      = __slam_tag_unix_group_suser(uid, gid);
	    tag_all_suser        = __slam_tag_unix_all_suser(uid);
	    tag_all_sgroup       = __slam_tag_unix_all_sgroup(gid);
	    tag_all_sgroup_auser = __slam_tag_unix_all_sgroup_auser(uid, gid);

	    /* rule ɲ */
	    __slam_add_rule(tag_all, tag_group);
	    __slam_add_rule(tag_all, tag_all_sgroup_auser);
	    __slam_add_rule(tag_all, tag_all_suser);
	    __slam_add_rule(tag_group, tag_user);
	    __slam_add_rule(tag_group, tag_group_suser);
	    __slam_add_rule(tag_all_sgroup_auser, tag_user);
	    __slam_add_rule(tag_all_sgroup_auser, tag_all_sgroup);
	    __slam_add_rule(tag_all_suser, tag_group_suser);
	    __slam_add_rule(tag_all_suser, tag_all_sgroup);
	    __slam_add_rule(tag_user, tag_root);
	    __slam_add_rule(tag_group_suser, tag_root);
	    __slam_add_rule(tag_all_sgroup, tag_root);

	    /*
	      ˤgroup tag ˿ user ޤޤ
	      rule ɲä롣
	    */
	    // gid б unix group  tag 
	    LOOP_ALL_LIST_NODE(__slam_tag_t, cur_tag,
			       tag_manager) {
		if (cur_tag->is_unix == FALSE) {
		    continue;
		}
		if (cur_tag->kind == UNIX_GROUP_SUB_USER &&
		    cur_tag->uid != uid &&
		    is_uid_member_of_gid(uid, cur_tag->gid)) {
		    // GROUP_SUB_USER  cur_tag 
		    // rule(cur_tag <= tag_user() ɲ
		    __slam_add_rule(cur_tag, tag_user);
		}
		else if (cur_tag->kind == UNIX_USER &&
			 cur_tag != tag_user) {
		    if (is_uid_member_of_gid(cur_tag->uid, gid)) {
			// USER  cur_tag 
			// rule(tag_group_suser <= cur_tag) ɲ
			__slam_add_rule(tag_group_suser, cur_tag);
		    }
		    else {
			__slam_add_rule(tag_all_sgroup, cur_tag);
		    }
		}
		else if (cur_tag->kind == UNIX_GROUP &&
			 cur_tag != tag_group) {
		    if (!is_uid_member_of_gid(uid, cur_tag->gid)) {
			__slam_add_rule(tag_all_suser, cur_tag);
		    }
		    else {
			__slam_add_rule(cur_tag, tag_user);
		    }
		}
		else if (cur_tag->kind == UNIX_ALL_SUB_GROUP &&
			 cur_tag != tag_all_sgroup) {
		    if (!is_uid_member_of_gid(uid, cur_tag->gid)) {
			__slam_add_rule(cur_tag, tag_user);
		    }
		    else {
			__slam_add_rule(tag_all_suser, cur_tag);
		    }
		}
		else if (cur_tag->kind == UNIX_ALL_SUB_USER &&
			 cur_tag != tag_all_suser ) {
		    if (cur_tag->uid != uid) {
			__slam_add_rule(cur_tag, tag_user);
		    }
		    if (!is_uid_member_of_gid(cur_tag->uid, gid)) {
			__slam_add_rule(cur_tag, tag_group);
		    }
		    else {
			__slam_add_rule(cur_tag, tag_all_sgroup);
		    }
		}
		else if (cur_tag->kind == UNIX_ALL_SUB_GROUP_SUB_USER &&
			 !is_uid_member_of_gid(uid, cur_tag->gid) &&
			 cur_tag->uid != uid) {
		    __slam_add_rule(cur_tag, tag_user);
		}
	    } LOOP_END;
	}
	else {
	    /* UID  GID ˴ޤޤʤ */

	    /* tag  */
	    tag_root             = __slam_tag_unix_root;
	    tag_user             = __slam_tag_unix_user(uid);
	    tag_group            = __slam_tag_unix_group(gid);
	    tag_all              = __slam_tag_unix_all;
	    tag_group_auser      = __slam_tag_unix_group_auser(uid, gid);
	    tag_all_suser        = __slam_tag_unix_all_suser(uid);
	    tag_all_sgroup       = __slam_tag_unix_all_sgroup(gid);
	    tag_all_sgroup_suser = __slam_tag_unix_all_sgroup_suser(uid, gid);

	    /* rule ɲ */
	    __slam_add_rule(tag_all, tag_group_auser);
	    __slam_add_rule(tag_all, tag_all_suser);
	    __slam_add_rule(tag_all, tag_all_sgroup);
	    __slam_add_rule(tag_group_auser, tag_user);
	    __slam_add_rule(tag_group_auser, tag_group);
	    __slam_add_rule(tag_all_suser, tag_group);
	    __slam_add_rule(tag_all_suser, tag_all_sgroup_suser);
	    __slam_add_rule(tag_all_sgroup, tag_user);
	    __slam_add_rule(tag_all_sgroup, tag_all_sgroup_suser);
	    __slam_add_rule(tag_user, tag_root);
	    __slam_add_rule(tag_group, tag_root);
	    __slam_add_rule(tag_all_sgroup_suser, tag_root);

	    /*
	      ˤgroup tag ˿ user ޤޤ
	      rule ɲä롣
	    */
	    // gid б unix group  tag 
	    LOOP_ALL_LIST_NODE(__slam_tag_t, cur_tag,
			       tag_manager) {
		if (cur_tag->is_unix == FALSE) {
		    continue;
		}
		if (cur_tag->kind == UNIX_USER &&
		    cur_tag != tag_user) {
		    if (cur_tag->uid != uid) {
			__slam_add_rule(tag_all_suser, cur_tag);
		    }
		    if (is_uid_member_of_gid(cur_tag->uid, gid)) {
			__slam_add_rule(tag_group, cur_tag);
		    }
		    else {
			__slam_add_rule(tag_all_sgroup, cur_tag);
			if (cur_tag->uid != uid) {
			    __slam_add_rule(tag_all_sgroup_suser, cur_tag);
			}
		    }
		}
		else if (cur_tag->kind == UNIX_GROUP &&
			 cur_tag != tag_group) {
		    if (is_uid_member_of_gid(uid, cur_tag->gid)) {
			__slam_add_rule(cur_tag, tag_user);
		    }
		    else {
			__slam_add_rule(tag_all_suser, cur_tag);
		    }
		}
		else if (cur_tag->kind == UNIX_GROUP_SUB_USER) {
		    if (cur_tag->uid != uid &&
			is_uid_member_of_gid(uid, cur_tag->gid)) {
			__slam_add_rule(cur_tag, tag_user);
		    }
		}
		else if (cur_tag->kind == UNIX_ALL_SUB_USER &&
			 cur_tag != tag_all_suser) {
		    if (cur_tag->uid != uid) {
			__slam_add_rule(cur_tag, tag_user);
		    }
		    if (is_uid_member_of_gid(cur_tag->uid, gid)) {
			__slam_add_rule(cur_tag, tag_all_sgroup);
		    }
		    else {
			__slam_add_rule(cur_tag, tag_group);
		    }
		}
		else if (cur_tag->kind == UNIX_ALL_SUB_GROUP &&
			 cur_tag != tag_all_sgroup) {
		    if (is_uid_member_of_gid(uid, cur_tag->gid)) {
			__slam_add_rule(tag_all_suser, cur_tag);
		    }
		    else {
			__slam_add_rule(cur_tag, tag_user);
		    }
		}
	    } LOOP_END;
	}
    }

    ret_tag = NULL;
    /* ͤ */
    if (is_uid_member_of_gid(uid, gid)) {
	/* UID  GID ˴ޤޤ */
	if (mode & S_IRUSR)
	    if (mode & S_IRGRP)
		if (mode & S_IROTH)	/* rrr */
		    ret_tag = __slam_tag_unix_all;
		else			/* rr- */
		    ret_tag = __slam_tag_unix_group(gid);
	    else
		if (mode & S_IROTH)	/* r-r */
		    ret_tag = __slam_tag_unix_all_sgroup_auser(uid, gid);
		else			/* r-- */
		    ret_tag = __slam_tag_unix_user(uid);
	else
	    if (mode & S_IRGRP)
		if (mode & S_IROTH)	/* -rr */
		    ret_tag = __slam_tag_unix_all_suser(uid);
		else			/* -r- */
		    ret_tag = __slam_tag_unix_group_suser(uid, gid);
	    else
		if (mode & S_IROTH)	/* --r */
		    ret_tag = __slam_tag_unix_all_sgroup(gid);
		else			/* --- */
		    ret_tag = __slam_tag_unix_root;
    }
    else {
	if (mode & S_IRUSR)
	    if (mode & S_IRGRP)
		if (mode & S_IROTH)	/* rrr */
		    ret_tag = __slam_tag_unix_all;
		else			/* rr- */
		    ret_tag = __slam_tag_unix_group_auser(uid, gid);
	    else
		if (mode & S_IROTH)	/* r-r */
		    ret_tag = __slam_tag_unix_all_sgroup(gid);
		else			/* r-- */
		    ret_tag = __slam_tag_unix_user(uid);
	else
	    if (mode & S_IRGRP)
		if (mode & S_IROTH)	/* -rr */
		    ret_tag = __slam_tag_unix_all_suser(uid);
		else			/* -r- */
		    ret_tag = __slam_tag_unix_group(gid);
	    else
		if (mode & S_IROTH)	/* --r */
		    ret_tag = __slam_tag_unix_all_sgroup_suser(uid, gid);
		else			/* --- */
		    ret_tag = __slam_tag_unix_root;
    }
    return ret_tag;
}

/*!
  user(UID)  tag 
 */
extern __slam_tag_t __slam_tag_unix_user(
    uid_t	uid)
{
    __slam_tag_t	ret_tag;
    char		*name;

    /* root ξ */
    if (uid == 0) {
	ret_tag = __slam_tag_unix_root;
    }
    else {
	name = malloc(sizeof(char) * 32);
	sprintf(name, "%sUser[%d]", SLAM_PREFIX, uid);
	assert(strlen(name) < 32);
	ret_tag = __slam_add_tag_copy(name);
	ret_tag->is_unix = TRUE;
	ret_tag->kind = UNIX_USER;
	ret_tag->uid = uid;
	free(name);
    }

    return ret_tag;
}

/*!
  group(GID)  tag 
 */
extern __slam_tag_t __slam_tag_unix_group(
    gid_t	gid)
{
    __slam_tag_t	ret_tag;
    char		*name;

    name = malloc(sizeof(char) * 32);
    sprintf(name, "%sGroup[%d]", SLAM_PREFIX, gid);
    assert(strlen(name) < 32);
    ret_tag = __slam_add_tag_copy(name);
    ret_tag->is_unix = TRUE;
    ret_tag->kind = UNIX_GROUP;
    ret_tag->gid = gid;
    free(name);

    return ret_tag;
}

/*!
  mode : -r-  group  user ޤޤ
   tag 
 */
static __slam_tag_t __slam_tag_unix_group_suser(
    uid_t	uid,
    gid_t	gid)
{
    __slam_tag_t	ret_tag;
    char		*name;

    name = malloc(sizeof(char) * 64);
    sprintf(name, "%sG[%d]-U[%d]", SLAM_PREFIX, gid, uid);
    assert(strlen(name) < 64);
    ret_tag = __slam_add_tag_copy(name);
    ret_tag->is_unix = TRUE;
    ret_tag->kind = UNIX_GROUP_SUB_USER;
    ret_tag->uid = uid;
    ret_tag->gid = gid;
    free(name);

    return ret_tag;
}

/*!
  mode : rr-  group  user ޤޤʤ
   tag 
 */
static __slam_tag_t __slam_tag_unix_group_auser(
    uid_t	uid,
    gid_t	gid)
{
    __slam_tag_t	ret_tag;
    char		*name;

    name = malloc(sizeof(char) * 64);
    sprintf(name, "%sG[%d]+U[%d]", SLAM_PREFIX, gid, uid);
    assert(strlen(name) < 64);
    ret_tag = __slam_add_tag_copy(name);
    ret_tag->is_unix = TRUE;
    ret_tag->kind = UNIX_GROUP_ADD_USER;
    ret_tag->uid = uid;
    ret_tag->gid = gid;
    free(name);

    return ret_tag;
}

/*!
  mode : -rr ξ tag 
 */
static __slam_tag_t __slam_tag_unix_all_suser(
    uid_t	uid)
{
    __slam_tag_t	ret_tag;
    char		*name;

    name = malloc(sizeof(char) * 64);
    sprintf(name, "%sALL-U[%d]", SLAM_PREFIX, uid);
    assert(strlen(name) < 64);
    ret_tag = __slam_add_tag_copy(name);
    ret_tag->is_unix = TRUE;
    ret_tag->kind = UNIX_ALL_SUB_USER;
    ret_tag->uid = uid;
    free(name);

    return ret_tag;
}

/*!
  mode : r-r  group  user ޤޤʤ
  mode : --r  group  user ޤޤ
   tag 
 */
static __slam_tag_t __slam_tag_unix_all_sgroup(
    gid_t	gid)
{
    __slam_tag_t	ret_tag;
    char		*name;

    name = malloc(sizeof(char) * 32);
    sprintf(name, "%sAll-G[%d]", SLAM_PREFIX, gid);
    assert(strlen(name) < 32);
    ret_tag = __slam_add_tag_copy(name);
    ret_tag->is_unix = TRUE;
    ret_tag->kind = UNIX_ALL_SUB_GROUP;
    ret_tag->gid = gid;
    free(name);

    return ret_tag;
}

/*!
  mode : --r  group  user ޤޤ
   tag 
 */
static __slam_tag_t __slam_tag_unix_all_sgroup_suser(
    uid_t	uid,
    gid_t	gid)
{
    __slam_tag_t	ret_tag;
    char		*name;

    name = malloc(sizeof(char) * 64);
    sprintf(name, "%sAll-G[%d]-U[%d]", SLAM_PREFIX, gid, uid);
    assert(strlen(name) < 64);
    ret_tag = __slam_add_tag_copy(name);
    ret_tag->is_unix = TRUE;
    ret_tag->kind = UNIX_ALL_SUB_GROUP_SUB_USER;
    ret_tag->uid = uid;
    ret_tag->gid = gid;
    free(name);

    return ret_tag;
}

/*!
  mode : r-r  group  user ޤޤʤ
   tag 
 */
static __slam_tag_t __slam_tag_unix_all_sgroup_auser(
    uid_t	uid,
    gid_t	gid)
{
    __slam_tag_t	ret_tag;
    char		*name;

    name = malloc(sizeof(char) * 64);
    sprintf(name, "%sALL-G[%d]+U[%d]", SLAM_PREFIX, gid, uid);
    assert(strlen(name) < 64);
    ret_tag = __slam_add_tag_copy(name);
    ret_tag->is_unix = TRUE;
    ret_tag->kind = UNIX_ALL_SUB_GROUP_ADD_USER;
    ret_tag->uid = uid;
    ret_tag->gid = gid;
    free(name);

    return ret_tag;
}


/*!
 - stat(const char *path, struct stat *buf);
 - fstat(int filedes, struct stat *buf);
 - lstat(const char *path, struct stat *buf);
 */
/* stat() permissiontag */
extern __slam_tag_t __slam_tag_stat(
    const char	*path)
{
    struct stat status;
    
    if (stat(path, &status) < 0 ) {
	fprintf(stderr, "Error: cannot __slam_tag_stat(\"%s\");\n", path);
	abort();
    }

    return __slam_tag_unix(status.st_uid,
			   status.st_gid,
			   status.st_mode);
}

/* fstat() permissiontag */
extern __slam_tag_t __slam_tag_fstat(
    int		filedes)
{
    struct stat status;
    
    if (fstat(filedes, &status) < 0 ) {
	fprintf(stderr, "Error: cannot __slam_tag_fstat(%d);\n", filedes);
	abort();
    }

    return __slam_tag_unix(status.st_uid,
			   status.st_gid,
			   status.st_mode);
}

/* lstat() permissiontag */
extern __slam_tag_t __slam_tag_lstat(
    const char	*path)
{
    struct stat status;
    
    if (lstat(path, &status) < 0 ) {
	fprintf(stderr, "Error: cannot __slam_tag_lstat(\"%s\");\n", path);
	abort();
    }

    return __slam_tag_unix(status.st_uid,
			   status.st_gid,
			   status.st_mode);
}

/* ǥХå */
extern void __slam_tag_name_dump(
    FILE		*file,
    __slam_tag_t	tag)
{
    fprintf(file, "%s ", tag->name);
}

extern void __slam_tag_dump(
    FILE		*file,
    __slam_tag_t	tag)
{
#ifndef NDEBUG
    fprintf(file, "Tag   %3d: ", tag->id);
#else
    fprintf(file, "Tag      : ");
#endif
    __slam_tag_name_dump(file, tag);

    fprintf(file, "\tbig(");
    LOOP_ALL_LIST_NODE(__slam_tag_t, big_tag, tag->big_set) {
	__slam_tag_name_dump(file, big_tag);
    } LOOP_END;
    fprintf(file, ")");

    fprintf(file, "\tsmall(");
    LOOP_ALL_LIST_NODE(__slam_tag_t, small_tag, tag->small_set) {
	__slam_tag_name_dump(file, small_tag);
    } LOOP_END;
    fprintf(file, ")");

    fprintf(file, "\n");
}

extern void __slam_tag_all_dump(
    FILE		*file)
{
#ifndef NDEBUG_USE_GRAPH
    FILE *graph_file;
    __slam_boolean_t	is_first;
#endif

    fprintf(file, "------ All Tag Dump ------\n");
    LOOP_ALL_LIST_NODE(__slam_tag_t, tag, tag_manager) {
	__slam_tag_dump(file, tag);
    } LOOP_END;
    fprintf(file, "--------------------------\n");

#ifndef NDEBUG_USE_GRAPH
    graph_file = fopen("digraph_tag.dot", "w");
    fprintf(graph_file, "digraph tag_lattice {\n");

    LOOP_ALL_LIST_NODE(__slam_tag_t, tag, tag_manager) {
	if (tag->small_set->length == 0) {
	    continue;
	}
	fprintf(graph_file, "\"");
	__slam_tag_name_dump(graph_file, tag);
	fprintf(graph_file, "\" -> {\"");

	is_first = TRUE;
	LOOP_ALL_LIST_NODE(__slam_tag_t, target_tag,
			   tag->small_tags) {
	    if (target_tag == tag) {
		continue;
	    }
	    if (is_first) {
		is_first = FALSE;
	    }
	    else {
		fprintf(graph_file, " \"");
	    }
	    __slam_tag_name_dump(graph_file, target_tag);	    
	    fprintf(graph_file, "\" ");
	} LOOP_END;
	fprintf(graph_file, "};\n");
    } LOOP_END;
    fprintf(graph_file, "}\n");

    fclose(graph_file);
#endif
}

/*!
  
 */
static __slam_tag_t tag_new(
    char	*tag_name)
{
    __slam_tag_t	tag;

    tag = (__slam_tag_t)malloc(sizeof(__slam_tag_type));
    tag->name = tag_name;
#ifndef NDEBUG
    tag->id  = tag_counter;
#endif
    tag->big_set = __slam_list_new();
    tag->small_set = __slam_list_new();
#ifndef NDEBUG
    tag->small_tags = __slam_list_new();
#endif
    tag->kind = GENERAL;
    __slam_list_insert(tag_manager, tag);

    tag_hash_enter(tag_hash_table, tag_name, tag);

#ifndef NDEBUG
    tag_counter += 1;
#endif

    return tag;
}

/*!
  ä
 */
static void tag_destroy(
    __slam_tag_t	tag)
{
    __slam_list_destroy(tag->small_set);
    __slam_list_destroy(tag->big_set);
    free(tag);
    return;
}

/*!
  ϥåơ֥
 */
static tag_hash_table_t *tag_hash_create(
    void)
{
    tag_hash_table_t *hash_table;

    hash_table = (tag_hash_table_t *)calloc(HASH_TABLE_SIZE,
						sizeof(tag_hash_data_t *));
    if (hash_table == NULL) {
	printf("calloc error\n");
    }
    return hash_table;
}

/*!
  ϥåơ֥Ȥ򥯥ꥢ
 */
static void tag_hash_flush(
    tag_hash_table_t *hash_table)
{
    int i;
    tag_hash_data_t	*cur_hash_data;
    tag_hash_data_t	*next_hash_data;

    for(i = 0; i < HASH_TABLE_SIZE; i++) {
	cur_hash_data = hash_table[i];
	while (cur_hash_data) {
	    next_hash_data = cur_hash_data->next;
	    free(cur_hash_data);
	    cur_hash_data = next_hash_data;
	}
    }
    return;
}

/*!
  ϥåơ֥õ
 */
static void tag_hash_destroy(
    tag_hash_table_t *hash_table)
{
    tag_hash_flush(hash_table);
    free(hash_table);
    return;
}

/*!
  ϥåơ֥tagϿ
 */
static void tag_hash_enter(
    tag_hash_table_t	*hash_table,
    char		*tag_name,
    __slam_tag_t	tag)
{
    tag_hash_data_t	*new_hash_data;
    tag_hash_data_t	*cur_hash_data;
    tag_hash_data_t	*prev_hash_data;
    
    cur_hash_data = hash_table[(int)(*tag_name) % HASH_TABLE_SIZE];
    prev_hash_data = NULL;
    while (cur_hash_data) {
	if (strcmp(cur_hash_data->tag_name, tag_name) == 0) {
	    /* ϿѤ */
	    return;
	}
	prev_hash_data = cur_hash_data;
	cur_hash_data = cur_hash_data->next;
    }

    /* ơ֥ */
    new_hash_data = (tag_hash_data_t *)malloc(sizeof(tag_hash_data_t));
    new_hash_data->tag_name = tag_name;
    new_hash_data->tag = tag;
    new_hash_data->next = cur_hash_data;

    if (prev_hash_data == NULL) {
	hash_table[(int)(*tag_name) % HASH_TABLE_SIZE] = new_hash_data;
    }
    else {
	prev_hash_data->next = new_hash_data;
    }
    return;
}

/*!
  ϥåơ֥뤫tag õ
 */
static __slam_tag_t tag_hash_find(
    tag_hash_table_t	*hash_table,
    char		*name)
{
    tag_hash_data_t	*hash_data;

    hash_data = hash_table[(int)(*name) % HASH_TABLE_SIZE];

    while (hash_data) {
	if (strcmp(hash_data->tag_name, name) == 0) {
	    /* ȯ */
	    return hash_data->tag;
	}
	else {
	    /* õ */
	    hash_data = hash_data->next;
	}
    }
    return NULL;
}

/*!
  ϥåơ֥
 */
static tag_unix_hash_table_t *tag_unix_hash_create(
    void)
{
    tag_unix_hash_table_t *hash_table;

    hash_table = (tag_unix_hash_table_t *)
	calloc(HASH_TABLE_SIZE, sizeof(tag_unix_hash_data_t *));
    if (hash_table == NULL) {
	printf("calloc error\n");
    }
    return hash_table;
}

/*!
  ϥåơ֥Ȥ򥯥ꥢ
 */
static void tag_unix_hash_flush(
    tag_unix_hash_table_t *hash_table)
{
    int i;
    tag_unix_hash_data_t	*cur_hash_data;
    tag_unix_hash_data_t	*next_hash_data;

    for(i = 0; i < HASH_TABLE_SIZE; i++) {
	cur_hash_data = hash_table[i];
	while (cur_hash_data) {
	    next_hash_data = cur_hash_data->next;
	    free(cur_hash_data);
	    cur_hash_data = next_hash_data;
	}
    }
    return;
}

/*!
  ϥåơ֥ξõ
 */
static void tag_unix_hash_destroy(
    tag_unix_hash_table_t *hash_table)
{
    tag_unix_hash_flush(hash_table);
    free(hash_table);
    return;
}

/*!
  ϥåơ֥key(uid,gid)unix tagѤ
  Ǥ뤳ȤϿ
 */
static void tag_unix_hash_enter(
    tag_unix_hash_table_t	*hash_table,
    long long int		key,
    __slam_boolean_t		has_tags)
{
    tag_unix_hash_data_t	*new_hash_data;
    tag_unix_hash_data_t	*cur_hash_data;
    tag_unix_hash_data_t	*prev_hash_data;
    
    cur_hash_data = hash_table[key % HASH_TABLE_SIZE];
    prev_hash_data = NULL;
    while (cur_hash_data) {
	if (cur_hash_data->uid_gid < key) {
	    break;
	}
	else if (cur_hash_data->uid_gid == key) {
	    /* ϿѤ */
	    return;
	}
	prev_hash_data = cur_hash_data;
	cur_hash_data = cur_hash_data->next;
    }

    /* ơ֥ */
    new_hash_data = (tag_unix_hash_data_t *)malloc(sizeof(tag_unix_hash_data_t));
    new_hash_data->uid_gid = key;
    new_hash_data->has_tags = has_tags;
    new_hash_data->next = cur_hash_data;

    if (prev_hash_data == NULL) {
	hash_table[key % HASH_TABLE_SIZE] = new_hash_data;
    }
    else {
	prev_hash_data->next = new_hash_data;
    }
    return;
}

/*!
  ϥåơ֥key(uid,gid)unix tagѤ
  Ǥ뤳ȤϿ
 */
static __slam_boolean_t tag_unix_hash_find(
    tag_unix_hash_table_t	*hash_table,
    long long int		key)
{
    tag_unix_hash_data_t	*hash_data;
    __slam_boolean_t		ret_val;

    hash_data = hash_table[key % HASH_TABLE_SIZE];

    ret_val = FALSE;
    while (hash_data) {
	if (hash_data->uid_gid == key) {
	    /* ȯ */
	    ret_val = hash_data->has_tags;
	    break;
	}
	else if (hash_data->uid_gid < key) {
	    /* ϥåˤʤ */
	    break;
	}
	else {
	    /* õ */
	    hash_data = hash_data->next;
	}
    }
    return ret_val;
}

/*!
  gidΥ롼פuidΥ桼°Ƥ뤫Ĵ٤
 */
static __slam_boolean_t is_uid_member_of_gid(
    uid_t	uid,
    gid_t	gid)
{
    struct passwd	*pw;
    struct group	*gr;
    int	i;
    __slam_boolean_t	ret_val;

    pw = getpwuid(uid);
    if (pw == NULL) {
	fprintf(stderr, "no user(%d)\n", uid);
	abort();
    }

    gr = getgrgid(gid);
    if (gr == NULL) {
	fprintf(stderr, "no group(%d)\n", gid);
	abort();
    }

    if (pw->pw_gid == gr->gr_gid) {
	ret_val = TRUE;
    }
    else {
	ret_val = FALSE;
	for(i = 0; gr->gr_mem[i] != NULL; i++) {
	    if (strcmp(pw->pw_name, gr->gr_mem[i]) == 0) {
		ret_val = TRUE;
		break;
	    }
	}
    }
    return ret_val;
}

/*!
  <uid, gid> Ȥ߹碌 unix tag 
  ȤȤϿ뤷TRUE֤
  ϿѤߤǤСϿFALSE֤
 */
static __slam_boolean_t tag_unix_add_to_hash(
    uid_t	uid,
    gid_t	gid)
{
    long long int	key;
    __slam_boolean_t	has_tags;
    __slam_boolean_t	ret_val;

    key = ((long long int)gid << 32) + uid;

    has_tags = tag_unix_hash_find(tag_unix_hash_table, key);

    if (!has_tags) {
	tag_unix_hash_enter(tag_unix_hash_table, key, TRUE);
	ret_val = TRUE;
    }
    else {
	ret_val = FALSE;
    }
    return ret_val;
}
